/***************************************************************************
 *   Copyright (C) 2009 by Arthur Renato Mello <arthur@mandriva.com        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License as        *
 *   published by the Free Software Foundation; either version 2 of        *
 *   the License or (at your option) version 3 or any later version        *
 *   accepted by the membership of KDE e.V. (or its successor approved     *
 *   by the membership of KDE e.V.), which shall act as a proxy            *
 *   defined in Section 14 of version 3 of the license.                    *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "timeofthedaywallpaper.h"

#include <QFileInfo>
#include <QImage>
#include <QPainter>
#include <QDomDocument>
#include <QDebug>

#include <Plasma/Animator>
#include <Plasma/DataEngineManager>
#include <Plasma/Theme>

K_EXPORT_PLASMA_WALLPAPER(timeoftheday, TimeOfTheDayWallpaper)

bool operator<(const QSize &a, const QSize &b)
{
    float aW = a.width();
    float aH = a.height();
    float bW = b.width();
    float bH = b.height();

    return (aW / aH) < (bW / bH);
}

TimeOfTheDayWallpaper::TimeOfTheDayWallpaper(QObject * parent, const QVariantList & args )
    : Plasma::Wallpaper(parent, args),
    m_baseSecs(0),
    m_wallpaperPackage(0),
    m_transitionFrames(50),
    m_transitionDuration(100),
    m_configWidget(0)
{
    connect(&m_timer, SIGNAL(timeout()), this, SLOT(nextWallpaper()));
    connect(this, SIGNAL(renderCompleted(QImage)), this, SLOT(updateBackground(QImage)));
    m_configUi = new Ui::TimeOfTheDayConfig();
}

TimeOfTheDayWallpaper::~TimeOfTheDayWallpaper()
{
}

void TimeOfTheDayWallpaper::init(const KConfigGroup & config)
{
    m_timer.stop();
    calculateGeometry();

    m_xmlFileName = config.readEntry("timeofthedayxml", "/usr/share/mdk/backgrounds/Mandriva.xml");
    m_color = config.readEntry("wallpapercolor", QColor(56, 111, 150));
    m_resizeMethod = (ResizeMethod)config.readEntry("wallpaperposition", (int)ScaledResize);

    parseXmlFile();
}

void TimeOfTheDayWallpaper::save(KConfigGroup & config)
{
    config.writeEntry("timeofthedayxml", m_xmlFileName);
    config.writeEntry("wallpaperposition", (int)m_resizeMethod);
    config.writeEntry("wallpapercolor", m_color);
}

void TimeOfTheDayWallpaper::configWidgetDestroyed()
{
    m_configWidget = 0;
}

QWidget * TimeOfTheDayWallpaper::createConfigurationInterface(QWidget * parent)
{
    m_configWidget = new QWidget(parent);
    connect(m_configWidget, SIGNAL(destroyed(QObject*)), this, SLOT(configWidgetDestroyed()));

    m_configUi->setupUi(m_configWidget);

    m_configUi->m_resizeMethod->addItem(i18n("Scaled & Cropped"), ScaledAndCroppedResize);
    m_configUi->m_resizeMethod->addItem(i18n("Scaled"), ScaledResize);
    m_configUi->m_resizeMethod->addItem(i18n("Scaled, keep proportions"), MaxpectResize);
    m_configUi->m_resizeMethod->addItem(i18n("Centered"), CenteredResize);
    m_configUi->m_resizeMethod->addItem(i18n("Tiled"), TiledResize);
    m_configUi->m_resizeMethod->addItem(i18n("Center Tiled"), CenterTiledResize);

    for (int i = 0; i < m_configUi->m_resizeMethod->count(); ++i) {
        if (m_resizeMethod == m_configUi->m_resizeMethod->itemData(i).value<int>()) {
            m_configUi->m_resizeMethod->setCurrentIndex(i);
            break;
        }
    }

    connect(m_configUi->m_resizeMethod, SIGNAL(currentIndexChanged(int)), this, SLOT(positioningChanged(int)));

    m_configUi->m_color->setColor(m_color);
    connect(m_configUi->m_color, SIGNAL(changed(const QColor&)), this, SLOT(colorChanged(const QColor&)));

    m_configUi->m_urlRequester->setFilter("*.xml");
    m_configUi->m_urlRequester->setUrl(KUrl(m_xmlFileName));
    connect(m_configUi->m_urlRequester, SIGNAL(returnPressed()), this, SLOT(xmlFileChanged()));
    connect(m_configUi->m_urlRequester, SIGNAL(urlSelected(const KUrl&)), this, SLOT(xmlFileChanged()));

    connect(this, SIGNAL(settingsChanged(bool)), parent, SLOT(settingsChanged(bool)));

    return m_configWidget;
}

void TimeOfTheDayWallpaper::startWallpaperTransition()
{
    m_timer.stop();

    if (m_events.isEmpty()) {
        m_pixmap = QPixmap();
        emit update(boundingRect());
    } else {
        nextWallpaper();
    }
}

void TimeOfTheDayWallpaper::nextWallpaper()
{
    if (m_events.isEmpty()) {
        return;
    }

    m_currentTime = QTime::currentTime();

    QTime keyEvent(0,0,0);
    QList<QTime> times = m_events.keys();

    foreach (const QTime &time, times) {
        if (time <= m_currentTime) {
            keyEvent = time;
        } else {
            break;
        }
    }

    if (!m_events.contains(keyEvent)) {
        keyEvent = m_events.keys().last();
    }

    TimeOfTheDayEvent currentEvent = m_events[keyEvent];

    QSize keySize = chooseAspectRatio(currentEvent.fromFiles->keys());
    QString wallpaper = currentEvent.fromFiles->value(keySize);

    m_transitionFrames = 10;
    m_transitionDuration = 50;

    setWallpaper(wallpaper);

    if (currentEvent.transition) {
        QSize toKeySize = chooseAspectRatio(currentEvent.toFiles->keys());
        QString toWallpaper = currentEvent.toFiles->value(toKeySize);

        int duration = currentEvent.duration - keyEvent.secsTo(m_currentTime);

        m_transitionFrames = duration / 2;
        m_transitionDuration = duration * 1000;
        m_transitionWallpaper = toWallpaper;

        QTimer::singleShot(200, this, SLOT(setTransitionWallpaper()));
    }

    int duration = currentEvent.duration - keyEvent.secsTo(m_currentTime);

    if (duration >= 0) {
        m_timer.start(duration * 1000);
    }
}

void TimeOfTheDayWallpaper::setTransitionWallpaper()
{
    QFileInfo info(m_transitionWallpaper);
    m_previousModified = info.lastModified();

    renderWallpaper(m_transitionWallpaper);
}

void TimeOfTheDayWallpaper::setWallpaper(const QString &wallpaper)
{
    //if (wallpaper == m_currentWallpaper) {
    //    QFileInfo info(m_currentWallpaper);
    //    if (m_previousModified == info.lastModified()) {
    //        return;
    //    }
    //}

    m_currentWallpaper = wallpaper;
    QFileInfo info(wallpaper);
    m_previousModified = info.lastModified();

    renderWallpaper(wallpaper);
}

void TimeOfTheDayWallpaper::calculateGeometry()
{
    m_size = boundingRect().size().toSize();
}

void TimeOfTheDayWallpaper::paint(QPainter * painter, const QRectF & exposedRect)
{
    if (m_size != boundingRect().size().toSize()) {
        calculateGeometry();

        if (!m_size.isEmpty() && !m_img.isEmpty()) {
            renderWallpaper();
            return;
        }
    }

    if (m_pixmap.isNull()) {
        painter->fillRect(exposedRect, QBrush(m_color));
        return;
    }

    if (painter->worldMatrix() == QMatrix()) {
        painter->resetTransform();
    }

    painter->setCompositionMode(QPainter::CompositionMode_Source);
    painter->drawPixmap(exposedRect, m_pixmap, exposedRect.translated(-boundingRect().topLeft()));

    if (!m_oldFadedPixmap.isNull()) {
        painter->setCompositionMode(QPainter::CompositionMode_SourceAtop);
        painter->drawPixmap(exposedRect, m_oldFadedPixmap, exposedRect.translated(-boundingRect().topLeft()));
    }
}

void TimeOfTheDayWallpaper::colorChanged(const QColor& color)
{
    m_color = color;
    emit settingsChanged(true);
}

void TimeOfTheDayWallpaper::positioningChanged(int index)
{
    m_resizeMethod = (ResizeMethod)m_configUi->m_resizeMethod->itemData(index).value<int>();
    setResizeMethodHint(m_resizeMethod);
    emit settingsChanged(true);
}

void TimeOfTheDayWallpaper::xmlFileChanged()
{
    KUrl url = m_configUi->m_urlRequester->url();
    if (url.isLocalFile()) {
        m_xmlFileName = m_configUi->m_urlRequester->url().toLocalFile();
        if (!m_xmlFileName.isEmpty()) {
            parseXmlFile();
        }
        emit settingsChanged(true);
    }
}

void TimeOfTheDayWallpaper::renderWallpaper(const QString& image)
{
    if (!image.isEmpty()) {
        m_img = image;
    }

    if (m_img.isEmpty()) {
        return;
    }

    render(m_img, m_size, m_resizeMethod, m_color);
}

void TimeOfTheDayWallpaper::updateBackground(const QImage &img)
{
    m_oldPixmap = m_pixmap;
    m_oldFadedPixmap = m_oldPixmap;
    m_pixmap = QPixmap::fromImage(img);
    int code = qrand();
    img.save(QString("/tmp/img-%1.png").arg(code));

    if (!m_oldPixmap.isNull()) {
        Plasma::Animator::self()->customAnimation(m_transitionFrames,
                                                  m_transitionDuration,
                                                  Plasma::Animator::LinearCurve,
                                                  this,
                                                  "updateFadedImage");
    } else {
        emit update(boundingRect());
    }
}

void TimeOfTheDayWallpaper::updateFadedImage(qreal frame)
{
    //If we are done, delete the pixmaps and don't draw.
    if (qFuzzyCompare(frame, qreal(1.0))) {
        m_oldFadedPixmap = QPixmap();
        m_oldPixmap = QPixmap();
        emit update(boundingRect());
        return;
    }

    //Create the faded image.
    m_oldFadedPixmap.fill(Qt::transparent);

    QPainter p;
    p.begin(&m_oldFadedPixmap);
    p.drawPixmap(0, 0, m_oldPixmap);
    p.setCompositionMode(QPainter::CompositionMode_DestinationIn);
    p.fillRect(m_oldFadedPixmap.rect(), QColor(0, 0, 0, 254 * (1-frame)));
    p.end();

    emit update(boundingRect());
}

void TimeOfTheDayWallpaper::parseXmlFile()
{
    if (m_xmlFileName.isEmpty()) {
        return;
    }

    // read the XMLfile
    QDomDocument xmldoc = QDomDocument(m_xmlFileName);
    QFile file(m_xmlFileName);
    if (!file.open(QIODevice::ReadOnly)) {
        return;
    }

    if (!xmldoc.setContent(&file)) {
        file.close();
        return;
    }

    file.close();

    m_events.clear();
    QDomElement docElem = xmldoc.documentElement();
    QDomNode n = docElem.firstChild();

    while(!n.isNull()) {
        QDomElement e = n.toElement(); // try to convert the node to an element.
        if (!e.isNull()) {
            if (e.tagName() == "starttime") {
                setStartTime(e);
            } else if (e.tagName() == "transition") {
                addTransition(e);
            } else if (e.tagName() == "static") {
                addStatic(e);
            }
        }
        n = n.nextSibling();
    }

    QTimer::singleShot(0, this, SLOT(startWallpaperTransition()));
}

void TimeOfTheDayWallpaper::setStartTime(QDomElement docElem)
{
    int hour = 0;
    int minute = 0;

    QDomNode n = docElem.firstChild();
    while (!n.isNull()) {
        QDomElement e = n.toElement();
        if (!e.isNull()) {
            if (e.tagName() == "hour") {
                hour = e.text().toInt();
            } else if (e.tagName() == "minute") {
                minute = e.text().toInt();
            }
        }
        n = n.nextSibling();
    }

    m_baseSecs = (hour * 60 * 60) + (minute * 60);
}

void TimeOfTheDayWallpaper::addStatic(QDomElement docElem)
{
    int duration = 0;
    QMap<QSize, QString> *files = new QMap<QSize, QString>();

    QDomNode n = docElem.firstChild();
    while (!n.isNull()) {
        QDomElement e = n.toElement();
        if (!e.isNull()) {
            if (e.tagName() == "duration") {
                duration = e.text().toFloat();

            } else if (e.tagName() == "file") {
                QDomNode f = e.firstChild();

                while (!f.isNull()) {
                    QDomElement s = f.toElement();

                    if (s.tagName() == "size") {
                        int w = s.attribute("width", "0").toInt();
                        int h = s.attribute("height", "0").toInt();

                        files->insert(QSize(w, h), s.text());
                    }

                    f = f.nextSibling();
                }
            }
        }

        n = n.nextSibling();
    }

    TimeOfTheDayEvent e = {false, files, 0, duration};

    QTime startTime(0,0,0);
    startTime = startTime.addSecs(m_baseSecs);
    m_baseSecs += duration;

    m_events.insert(startTime, e);
}

void TimeOfTheDayWallpaper::addTransition(QDomElement docElem)
{
    int duration = 0;
    QMap<QSize, QString> *fromFiles  = new QMap<QSize, QString>();
    QMap<QSize, QString> *toFiles  = new QMap<QSize, QString>();

    QDomNode n = docElem.firstChild();
    while (!n.isNull()) {
        QDomElement e = n.toElement();
        if (!e.isNull()) {
            if (e.tagName() == "duration") {
                duration = e.text().toFloat();

            } else if (e.tagName() == "from") {
                QDomNode f = e.firstChild();

                while (!f.isNull()) {
                    QDomElement s = f.toElement();

                    if (s.tagName() == "size") {
                        int w = s.attribute("width", "0").toInt();
                        int h = s.attribute("height", "0").toInt();

                        fromFiles->insert(QSize(w, h), s.text());
                    }

                    f = f.nextSibling();
                }

            } else if (e.tagName() == "to") {
                QDomNode f = e.firstChild();

                while (!f.isNull()) {
                    QDomElement s = f.toElement();

                    if (s.tagName() == "size") {
                        int w = s.attribute("width", "0").toInt();
                        int h = s.attribute("height", "0").toInt();

                        toFiles->insert(QSize(w, h), s.text());
                    }

                    f = f.nextSibling();
                }
            }
        }

        n = n.nextSibling();
    }

    TimeOfTheDayEvent e = {true, fromFiles, toFiles, duration};

    QTime startTime(0,0,0);
    startTime = startTime.addSecs(m_baseSecs);
    m_baseSecs += duration;

    m_events.insert(startTime, e);
}

QSize TimeOfTheDayWallpaper::chooseAspectRatio(QList<QSize> list)
{
    if (list.isEmpty()) {
        return QSize(0, 0);
    }

    if (list.contains(m_size)) {
        return m_size;
    }

    QSize choose(0,0);
    float diff = -1;
    float currentW = m_size.width();
    float currentH = m_size.height();
    float currentAspectRatio = currentW / currentH;

    foreach (QSize s, list) {
        float sW = m_size.width();
        float sH = m_size.height();
        float sAspectRatio = sW/ sH;

        float newDiff = currentAspectRatio - sAspectRatio;
        if (newDiff < 0) {
            newDiff *= -1;
        }

        if (diff < 0 || diff > newDiff) {
            choose = s;
            diff = newDiff;
        }
    }

    return choose;
}

#include "timeofthedaywallpaper.moc"
